#[cfg(feature = "serde")]
#[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
pub mod serde;

use basen::Base;
use uuid::Uuid;

const BASE58_FLICKR: Base<58> = base(b"123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ");
const BASE62: Base<62> = base(b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789");
const BASE64: Base<64> = base(b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/");
const BASE64_URL: Base<64> =
    base(b"ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_");
const BASE90_COOKIE: Base<90> = base(
    b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!#$%&'()*+-./:<=>?@[]^_`{|}~",
);

const fn base<const N: usize>(chars: &[u8; N]) -> Base<N> {
    match Base::new(chars) {
        Some(val) => val,
        None => panic!(),
    }
}

/// # Base58 encoding
///
/// Charset: `123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz`
pub trait Base58: Sized {
    fn encode_base58(&self) -> String;
    fn decode_base58(input: &str) -> Option<Self>;
}

/// # Base58 (Flickr) encoding
///
/// Charset: `123456789abcdefghijkmnopqrstuvwxyzABCDEFGHJKLMNPQRSTUVWXYZ`
///
/// Compatible with [short-uuid](https://www.npmjs.com/package/short-uuid)
pub trait Base58Flickr: Sized {
    fn encode_base58_flickr(&self) -> String;
    fn decode_base58_flickr(input: &str) -> Option<Self>;
}

/// # Base62 encoding
///
/// Charset: `ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789`
pub trait Base62: Sized {
    fn encode_base62(&self) -> String;
    fn decode_base62(input: &str) -> Option<Self>;
}

/// # Base64 encoding
///
/// Charset: `ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/`
pub trait Base64: Sized {
    fn encode_base64(&self) -> String;
    fn decode_base64(input: &str) -> Option<Self>;
}

/// # Base64 (URL) encoding
///
/// Charset: `ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789-_`
pub trait Base64Url: Sized {
    fn encode_base64_url(&self) -> String;
    fn decode_base64_url(input: &str) -> Option<Self>;
}

/// # Base90 (Cookie) encoding
///
/// Charset: ``0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ!#$%&'()*+-./:<=>?@[]^_`{|}~``
///
/// Compatible with [short-uuid](https://www.npmjs.com/package/short-uuid)
pub trait Base90Cookie: Sized {
    fn encode_base90_cookie(&self) -> String;
    fn decode_base90_cookie(input: &str) -> Option<Self>;
}

impl Base58 for Uuid {
    fn encode_base58(&self) -> String {
        basen::BASE58.encode_const_len(self.as_bytes())
    }

    fn decode_base58(input: &str) -> Option<Self> {
        basen::BASE58.decode_const_len(input).map(Self::from_bytes)
    }
}

impl Base58Flickr for Uuid {
    fn encode_base58_flickr(&self) -> String {
        BASE58_FLICKR.encode_const_len(self.as_bytes())
    }

    fn decode_base58_flickr(input: &str) -> Option<Self> {
        BASE58_FLICKR.decode_const_len(input).map(Self::from_bytes)
    }
}

impl Base62 for Uuid {
    fn encode_base62(&self) -> String {
        BASE62.encode_const_len(self.as_bytes())
    }

    fn decode_base62(input: &str) -> Option<Self> {
        BASE62.decode_const_len(input).map(Self::from_bytes)
    }
}

impl Base64 for Uuid {
    fn encode_base64(&self) -> String {
        BASE64.encode_const_len(self.as_bytes())
    }

    fn decode_base64(input: &str) -> Option<Self> {
        BASE64.decode_const_len(input).map(Self::from_bytes)
    }
}

impl Base64Url for Uuid {
    fn encode_base64_url(&self) -> String {
        BASE64_URL.encode_const_len(self.as_bytes())
    }

    fn decode_base64_url(input: &str) -> Option<Self> {
        BASE64_URL.decode_const_len(input).map(Self::from_bytes)
    }
}

impl Base90Cookie for Uuid {
    fn encode_base90_cookie(&self) -> String {
        BASE90_COOKIE.encode_const_len(self.as_bytes())
    }

    fn decode_base90_cookie(input: &str) -> Option<Self> {
        BASE90_COOKIE.decode_const_len(input).map(Self::from_bytes)
    }
}
