use serde::{Deserializer, Serializer};
use uuid::Uuid;

use crate::{Base58, Base58Flickr, Base62, Base64, Base64Url, Base90Cookie};

macro_rules! serde_impl {
    ($name:literal, $mname:ident, $enc:ident, $dec:ident) => {
        pub mod $mname {
            use super::*;

            pub fn serialize<S: Serializer>(
                value: &Uuid,
                serializer: S,
            ) -> Result<S::Ok, S::Error> {
                serializer.serialize_str(&value.$enc())
            }

            pub fn deserialize<'de, D: Deserializer<'de>>(
                deserializer: D,
            ) -> Result<Uuid, D::Error> {
                struct Visitor;

                impl<'de> serde::de::Visitor<'de> for Visitor {
                    type Value = Uuid;

                    fn expecting(&self, formatter: &mut std::fmt::Formatter) -> std::fmt::Result {
                        formatter.write_str($name)?;
                        formatter.write_str("-formatted UUID")
                    }

                    fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
                    where
                        E: serde::de::Error,
                    {
                        Uuid::$dec(v).ok_or(E::custom("could not parse uuid"))
                    }
                }
                deserializer.deserialize_str(Visitor)
            }
        }
    };
}

serde_impl!("base58", base58, encode_base58, decode_base58);
serde_impl!(
    "base58_flickr",
    base58_flickr,
    encode_base58_flickr,
    decode_base58_flickr
);
serde_impl!("base62", base62, encode_base62, decode_base62);
serde_impl!("base64", base64, encode_base64, decode_base64);
serde_impl!(
    "base64_url",
    base64_url,
    encode_base64_url,
    decode_base64_url
);
serde_impl!(
    "base90_cookie",
    base90_cookie,
    encode_base90_cookie,
    decode_base90_cookie
);
